//
//  OCRSDK.swift
//  OCR-SDK
//
//  Created by itsol on 5/18/20.
//  Copyright © 2020 itsol. All rights reserved.
//

import Foundation
import UIKit

public class SB_KYC_SDK {
    public static let shared = SB_KYC_SDK()
    public typealias CompletionHandle = (_ data: [String: Any]?, _ cardImage: Data?, _ faceImage: Data?) -> Void
    public typealias CompletionHandleFace = (_ data: Bool) -> Void
    let viewController = SBKTutorialVC(URLToken: "http://sdk.sb.gotai.ml")
    
    //Khởi tạo framework
    //input: viewCurrent, sdkConfig
    public func initSDK(viewCurrent: UIViewController, sdkConfig: [String:Any], completion: @escaping CompletionHandle) {
        DispatchQueue.main.async {
            Global.language = sdkConfig["language"] as? String ?? Global.language
            Global.url_font = sdkConfig["url_font"] as? String ?? Global.url_font
            Global.url_back = sdkConfig["url_back"] as? String ?? Global.url_back
            Global.url_face = sdkConfig["url_face"] as? String ?? Global.url_face
            Global.header = sdkConfig["header"] as? [String: Any] ?? Global.header
            Global.typeFace = sdkConfig["type_face"] as? String ?? Global.typeFace
            Global.colorConfig = sdkConfig["color"] as? String ?? Global.colorConfig
            Global.frontConfig = sdkConfig["front"] as? String ?? Global.frontConfig
            Global.copyright = sdkConfig["copyright"] as? String ?? Global.copyright
            Global.step = sdkConfig["step"] as? String ?? Global.step
            Global.colorButtonNext = sdkConfig["colorBackgroundButtonNext"] as? String ?? Global.colorButtonNext
            Global.colorButtonBack = sdkConfig["colorBackgroundButtonBack"] as? String ?? Global.colorButtonBack
            Global.colorTextButtonNext = sdkConfig["colorTextButtonNext"] as? String ?? Global.colorTextButtonNext
            Global.colorTextButtonBack = sdkConfig["colorTextButtonBack"] as? String ?? Global.colorTextButtonBack
            Global.colorTextPreview = sdkConfig["colorTextPreview"] as? String ?? Global.colorTextPreview
            let ratio = sdkConfig["ratioPass"] as? Int ?? 70
            Global.ratioPass = Float(ratio)
            Global.useFileCer = sdkConfig[""] as? Bool ?? Global.useFileCer

            self.viewController.completionSuccessTutorial = { [weak self] data, cardImage, faceImage in
                guard self != nil else { return }
                completion(data, cardImage, faceImage)
            }
            viewCurrent.navigationController?.pushViewController(self.viewController, animated: true)
            viewCurrent.navigationController?.setNavigationBarHidden(false, animated: false)
        }
    }
    
    //Validate card
    //input: UIImage
    //output: Int, 0: Image fake, 1: Image front, 2: Image back, -1: error
    public func eKycCardValid(image: CVPixelBuffer) -> Int {
        let resultVaidate = SBKValidateInput.shared.didOutput(pixelBuffer: image)
        
        switch resultVaidate {
        case .ERROR:
            return -1
        case .IMAGE_FAKE:
            return 0
        case .IMAGE_FRONT:
            return 1
        case .IMAGE_BACK:
            return 2
        }
    }
    
    //Validate face
    //input: UIImage
    public func eKycFaceValid(image: CVPixelBuffer, completion: @escaping CompletionHandleFace){
        SBKValidateInput.shared.validateFace(imageFace: image) {[weak self] data in
            guard self != nil else { return }
            completion(data)
        }
    }
    
    //Call API face
    //input: UIImage
    public func eKycFaceMatch(imageFace: CVPixelBuffer, imageCardFront: CVPixelBuffer, imageCardBack: CVPixelBuffer, completion: @escaping CompletionHandle) {
        SBKValidateInput.shared.validateFace(imageFace: imageFace) {[weak self] data in
            guard self != nil else { return }
            if data {
                Global.imageCard1 = SBKValidateInput.shared.convertCVPixelToUIImage(pixelBuffer: imageCardFront).pngData()
                Global.imageCard2 = SBKValidateInput.shared.convertCVPixelToUIImage(pixelBuffer: imageCardFront).pngData()
                let imageFaceUI = SBKValidateInput.shared.convertCVPixelToUIImage(pixelBuffer: imageFace).pngData()
                
                SBOCRRequest.shared.processFace(image: imageFaceUI! , pathURL: Global.url_face, idBack: "", idFront: "") {(errorCode,data) -> Void in
                    guard errorCode == nil else {
                        let error: [String: Any] = ["error": errorCode!]
                        completion(error, nil, nil)
                        return
                    }
                   completion(data, nil, nil)
                }
            } else {
                let error: [String: Any] = ["error": "image faild"]
                completion(error, nil, nil)
            }
        }
    }
}
