//
//  SBKValidateCardView.swift
//  OCR-SDK
//
//  Created by itsol on 8/25/20.
//  Copyright © 2020 itsol. All rights reserved.
//

import UIKit
import AVFoundation

public class SBKValidateCardView: UIView, AVCapturePhotoCaptureDelegate {
    let nibName = "SBKValidateCardView"
    var contentView:UIView?
    
    @IBOutlet public weak var viewCamera: UIView!
    @IBOutlet public weak var btnTakePhoto: UIButton!
    
    public var captureSession: AVCaptureSession = AVCaptureSession()
    public var stillImageOutput: AVCapturePhotoOutput =  AVCapturePhotoOutput()
    public var videoPreviewLayer: AVCaptureVideoPreviewLayer!
    public let videoDataOutput = AVCaptureVideoDataOutput()
    public var inputCamera: AVCaptureDeviceInput!
    
    var descriptionScreen: String = "Front of your personal card".localized()
    var checkScreen: Int = 1 // checkScreen = 1: screen front card, checkScreen = 2: screen front
    var idFront: String = ""
    var URLToken: String = ""
    var statusTakePhoto: Bool = true
    var statusValidateImage: ValidateCard = .ERROR
    var statusScreen: String = "vertical" //horizontal
    public var typeCamera: TypeCard = TypeCard.FRONT
    
    public var completionSuccessCardStep: (_ validate: ValidateCard?, _ data: Data?, _ permissionCamera: Bool?)->Void = {_,_,_  in}
    
    public var iconTakeCard: Data = UIImage(named: "iconCap", in: Bundle(for: SBKTutorialVC.self), compatibleWith: nil)!.pngData()!
    
    public required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
        commonInit()
    }
    
    public override init(frame: CGRect) {
        super.init(frame: frame)
        commonInit()
    }
    
    func commonInit() {
        guard let view = loadViewFromNib() else { return }
        view.frame = self.bounds
        self.addSubview(view)
        contentView = view
        
        AVCaptureDevice.requestAccess(for: .video, completionHandler: { (granted: Bool) in
            if !granted {
              self.completionSuccessCardStep(nil, nil, false)
                return
            }
        })
        
        if AVCaptureDevice.authorizationStatus(for: .video) ==  .denied {
            self.completionSuccessCardStep(nil, nil, false)
            return
        }
        self.loadCamera()
        
    }
    
    func loadViewFromNib() -> UIView? {
        let bundle = Bundle(for: type(of: self))
        let nib = UINib(nibName: nibName, bundle: bundle)
        return nib.instantiate(withOwner: self, options: nil).first as? UIView
    }
    
    public func setButtonImage(iconTakeCard: UIImage? = nil, iconFrameTopLeft: UIImage? = nil, iconFrameTopRight: UIImage? = nil, iconFrameBottomLeft: UIImage? = nil, iconFrameBottomRight: UIImage? = nil) {
           
        let fWidth = self.bounds.size.width
        let fHeight = self.bounds.size.height
        let squareWidth = fWidth/2
        let topLeft = CGPoint(x: fWidth/2-squareWidth/3, y: fHeight/4)
        
        if iconTakeCard != nil {
            self.btnTakePhoto.setImage(iconTakeCard, for: .normal)
        }
        
        if iconFrameTopLeft != nil {
            let ivLeftTop = UIImageView(image: iconFrameTopLeft)
            ivLeftTop.frame = CGRect(x: self.viewCamera.frame.origin.x + fWidth/20, y: topLeft.y - fHeight/20, width: 50, height: 50)
            self.viewCamera.addSubview(ivLeftTop)
        }
        
        if iconFrameTopRight != nil {
            let ivRightTop = UIImageView(image: iconFrameTopRight)
            ivRightTop.frame = CGRect(x: self.viewCamera.frame.origin.x + fWidth/20 + fWidth * 18 / 20 - 50, y: topLeft.y - fHeight/20, width: 50, height: 50)
            self.viewCamera.addSubview(ivRightTop)
        }
        
        if iconFrameBottomLeft != nil {
            let ivLeftBottom = UIImageView(image: iconFrameBottomLeft)
            ivLeftBottom.frame = CGRect(x: self.viewCamera.frame.origin.x + fWidth/20, y: topLeft.y - fHeight/20 + fWidth * 18 * 3 / 20 / 4 - 50, width: 50, height: 50)
            self.viewCamera.addSubview(ivLeftBottom)
        }
        
        if iconFrameBottomRight != nil {
             let ivRightBottom = UIImageView(image: iconFrameBottomRight)
            ivRightBottom.frame = CGRect(x: self.viewCamera.frame.origin.x + fWidth/20 + fWidth * 18 / 20 - 50, y: topLeft.y - fHeight/20 + fWidth * 18 * 3 / 20 / 4 - 50, width: 50, height: 50)
            self.viewCamera.addSubview(ivRightBottom)
        }
    }
    
    public func startCamera() {
        self.captureSession.startRunning()
    }
    
    public func stopCamera() {
        self.captureSession.stopRunning()
    }
       
    //Cài đặt máy ảnh
    func loadCamera() {
        captureSession.sessionPreset = .hd1920x1080
        
        if #available(iOS 11.1, *) {
            guard let backCamera = AVCaptureDevice.DiscoverySession(
                deviceTypes: [.builtInWideAngleCamera, .builtInDualCamera, .builtInTrueDepthCamera],
                mediaType: .video,
                position: .back).devices.first else {
                    print("Unable to access back camera!")
                    return
            }
           
           do {
               self.inputCamera = try AVCaptureDeviceInput(device: backCamera)

               if captureSession.canAddInput(self.inputCamera) {
                   captureSession.addInput(self.inputCamera)
               }
               if captureSession.canAddOutput(stillImageOutput) {
                   captureSession.addOutput(stillImageOutput)
                   setupLivePreview()
               }
           }
           catch let error  {
               print("Error Unable to initialize back camera:  \(error.localizedDescription)")
           }
           
           self.getCameraFrames()
           
           DispatchQueue.global(qos: .userInitiated).async {
               self.captureSession.startRunning()
           }
           
           DispatchQueue.main.async {
               self.videoPreviewLayer.frame = self.viewCamera.bounds
           }
        }
       }
       
       
       //set frame preview
       func setupLivePreview() {
           videoPreviewLayer = AVCaptureVideoPreviewLayer(session: captureSession)
           
           videoPreviewLayer.videoGravity = .resizeAspectFill
           videoPreviewLayer.connection?.videoOrientation = .portrait
           viewCamera.layer.addSublayer(videoPreviewLayer)
           
           if let viewWithTag = self.viewCamera.viewWithTag(11) {
               viewWithTag.removeFromSuperview()
           }
           
           if let viewWithTag = self.viewCamera.viewWithTag(22) {
               viewWithTag.removeFromSuperview()
           }
        let overlay = createOverlay(frame: self.bounds)
           overlay.tag = 11
           viewCamera.addSubview(overlay)
       }
       
       //set vùng đặt thẻ
       func createOverlay(frame: CGRect) -> UIView {
           let overlayView = UIView(frame: frame)
           overlayView.backgroundColor = UIColor.black.withAlphaComponent(0.6)
           let path = CGMutablePath()
           
            let fWidth = self.bounds.size.width
           let fHeight = self.bounds.size.height
           let squareWidth = fWidth/2
           let topLeft = CGPoint(x: fWidth/2-squareWidth/3, y: fHeight/4)
            path.addRoundedRect(in: CGRect(x: self.viewCamera.frame.origin.x + fWidth/20, y: topLeft.y - fHeight/20,
                width: fWidth * 18 / 20, height: fWidth * 18 * 3 / 20 / 4 ),
               cornerWidth: 0, cornerHeight: 0)

           path.addRect(CGRect(origin: .zero, size: overlayView.frame.size))
           let maskLayer = CAShapeLayer()
           maskLayer.backgroundColor = UIColor.black.cgColor
           maskLayer.path = path
           maskLayer.fillRule = .evenOdd
           overlayView.layer.mask = maskLayer
           overlayView.clipsToBounds = true

           return overlayView
       }

       //Xử lý data sâu khi chụp
       @available(iOS 11.0, *)
    public func photoOutput(_ output: AVCapturePhotoOutput, didFinishProcessingPhoto photo: AVCapturePhoto, error: Error?) {
           
           guard let imageData = photo.fileDataRepresentation()
               else { return }
           
//           self.captureSession.stopRunning()
//           self.captureSession.removeOutput(self.videoDataOutput)
//           self.captureSession.removeOutput(self.stillImageOutput)
        let cropImage = self.cropImage(image: UIImage(data: imageData)!, rect: CGRect(x: UIImage(data: imageData)!.size.width / 20, y: UIImage(data: imageData)!.size.height * 6 / 20 , width: self.viewCamera.frame.size.width * 18 / 20, height: self.viewCamera.frame.size.width * 18 * 3 / 20 / 4), scale: 1.0)
            self.completionSuccessCardStep(nil, cropImage!.pngData()!, nil)
       }
       
       //Xử lý ảnh hiển thị
       func cropImage(image: UIImage, rect: CGRect, scale: CGFloat) -> UIImage? {
           let imageCap = image
           
           let widthCrop = imageCap.size.width - imageCap.size.width / 10
           
           UIGraphicsBeginImageContextWithOptions(CGSize(width: widthCrop, height: widthCrop * 3 / 4), true, 0.0)
           image.draw(at: CGPoint(x: -rect.origin.x / scale, y: -rect.origin.y / scale))
           let croppedImage = UIGraphicsGetImageFromCurrentImageContext()
           UIGraphicsEndImageContext()
           return croppedImage
       }
       
       //Sự kiện chụp ảnh
       @IBAction func onCapturePhoto(_ sender: Any) {
        if (self.statusValidateImage == ValidateCard.IMAGE_FRONT && self.typeCamera == TypeCard.FRONT) || (self.statusValidateImage == .IMAGE_BACK && self.typeCamera == TypeCard.BACK) {
               if #available(iOS 11.0, *) {
                   let settings = AVCapturePhotoSettings(format: [AVVideoCodecKey: AVVideoCodecType.jpeg])
                   stillImageOutput.capturePhoto(with: settings, delegate: self)
               } else {
                   let settings = AVCapturePhotoSettings(format: [AVVideoCodecKey: AVVideoCodecJPEG])
                   stillImageOutput.capturePhoto(with: settings, delegate: self)
               }
           }
       }
       
       func resizeImageCard(pixelBuffer: CVPixelBuffer) -> CVPixelBuffer {
           let validateView = SBKValidateInput.shared
           let imageInput = validateView.convertCVPixelToUIImage(pixelBuffer: pixelBuffer)
           var imageOutput: UIImage?
           if self.statusScreen == "horizontal" {
               imageOutput = validateView.cropImageHorizontal(image: imageInput, rect: CGRect(x: imageInput.size.width * 1 / 10, y: imageInput.size.height * 3 / 20, width: imageInput.size.width * 8 / 10, height: imageInput.size.height * 9 / 10), scale: 1.0)
           } else {
               imageOutput = validateView.cropImage(image: imageInput, rect: CGRect(x: imageInput.size.width / 20, y: imageInput.size.height * 2 / 7, width: imageInput.size.width - imageInput.size.width/10, height: (imageInput.size.width - imageInput.size.width/10) * 3/4), scale: 1.0)
           }
           
           let ciimage = CIImage(image: imageOutput!)
            let eaglContext = EAGLContext(api: .openGLES2)
            let tmpcontext = CIContext(eaglContext: eaglContext!)
           let cgimage =  tmpcontext.createCGImage(ciimage!, from: ciimage!.extent)
           
           return validateView.convertCGImgeToCVPixelBuffer(forImage: cgimage!)!
       }
}

extension SBKValidateCardView: AVCaptureVideoDataOutputSampleBufferDelegate {
    private func getCameraFrames() {
        self.videoDataOutput.videoSettings = [(kCVPixelBufferPixelFormatTypeKey as NSString) : NSNumber(value: kCVPixelFormatType_32BGRA)] as [String : Any]
        self.videoDataOutput.alwaysDiscardsLateVideoFrames = true
        self.videoDataOutput.setSampleBufferDelegate(self, queue: DispatchQueue(label: "camera_frame_processing_queue"))
        if captureSession.canAddOutput(self.videoDataOutput) {
            self.captureSession.addOutput(self.videoDataOutput)
        }
        //self.captureSession.addOutput(self.videoDataOutput)
        guard let connection = self.videoDataOutput.connection(with: AVMediaType.video),
            connection.isVideoOrientationSupported else { return }
        connection.videoOrientation = .portrait
    }
    
    public func captureOutput( _ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        
        guard let imageFrameInput = CMSampleBufferGetImageBuffer(sampleBuffer) else {
            debugPrint("unable to get image from sample buffer")
            return
        }
        let validateImageCard = SBKValidateInput.shared.didOutput(pixelBuffer: self.resizeImageCard(pixelBuffer: imageFrameInput))
        DispatchQueue.main.async {
            self.completionSuccessCardStep(validateImageCard, nil, nil)
            
            self.statusValidateImage = validateImageCard
        }
    }
}
