//
//  CaptureFaceVC.swift
//  SSSSS
//
//  Created by itsol on 5/12/20.
//  Copyright © 2020 itsol. All rights reserved.
//

import UIKit
import AVFoundation

class SBKCaptureFaceVC: UIViewController, AVCapturePhotoCaptureDelegate {
    var captureSession: AVCaptureSession = AVCaptureSession()
    var stillImageOutput: AVCapturePhotoOutput =  AVCapturePhotoOutput()
    var videoPreviewLayer: AVCaptureVideoPreviewLayer!
    let videoDataOutput = AVCaptureVideoDataOutput()
    var idFront: String = ""
    var idBack: String = ""
    var URLToken: String = ""
    var statusTake: Bool = false
    
    @IBOutlet weak var lbDescription: UILabel!
    @IBOutlet weak var viewbaackground: UIView!
    @IBOutlet weak var lbCopyright: UILabel!
    @IBOutlet weak var viewOval: UIView!
    
    var completionSuccessFace: (_ data: [String: Any]?, _ cardImage: Data?, _ faceImage: Data?) -> Void = {_,_,_ in}
//    var validateFace: (Bool) -> Void = {_ in}
//    var checkFace: (Bool) -> Void = {_ in}
    
    override func viewDidLoad() {
        super.viewDidLoad()
        var image = UIImage(named: "back", in: Bundle(for: SBKTutorialVC.self), compatibleWith: nil)
        image = image?.withRenderingMode(.alwaysOriginal)
        
        let button = UIButton(type: .system)
        button.setImage(image, for: .normal)
        button.setTitle("Back".localized(), for: .normal)
        button.sizeToFit()
        button.imageEdgeInsets.left = -5
        button.setTitleColor(UIColor.black, for: .normal)
        button.addTarget(self, action: #selector(self.back), for: .touchUpInside)
        
        let newBackButton = UIBarButtonItem(customView: button)
        self.navigationItem.leftBarButtonItem = newBackButton
    }
    
    //Xử lý back
    @objc func back(sender: UIBarButtonItem) {
        for controller in self.navigationController!.viewControllers as Array {
            if controller.isKind(of: SBKCaptureCardVC.self) {
                self.navigationController!.popToViewController(controller, animated: true)
                break
            }
        }
    }
    
    func loadConfigUI() {
        self.lbCopyright.text = Global.copyright
        self.lbCopyright.textColor = UIColor.colorFromHexa(Global.colorConfig)
        self.lbDescription.textColor = UIColor.colorFromHexa(Global.colorConfig)
        if Global.frontConfig == "" {
            self.lbCopyright.font = UIFont(name: Global.frontConfig, size: 17)
            self.lbDescription.font = UIFont(name: Global.frontConfig, size: 17)
        }
    }

    //Chụp ảnh
    @IBAction func onCapture(_ sender: Any) {
        if self.statusTake {
            if #available(iOS 11.0, *) {
                let settings = AVCapturePhotoSettings(format: [AVVideoCodecKey: AVVideoCodecType.jpeg])
                stillImageOutput.capturePhoto(with: settings, delegate: self)
            } else {
                let settings = AVCapturePhotoSettings(format: [AVVideoCodecKey: AVVideoCodecJPEG])
                stillImageOutput.capturePhoto(with: settings, delegate: self)
            }
        }
    }
    
    init() {
        super.init(nibName: "SBKCaptureFaceVC", bundle: Bundle(for: SBKCaptureFaceVC.self))
    }
          
    required init?(coder aDecoder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    //Cài đặt camera
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        
        self.loadConfigUI()
        self.lbDescription.text = "Position your face in the oval".localized()
        
        captureSession.sessionPreset = .medium
           
        guard let backCamera = AVCaptureDevice.default(.builtInWideAngleCamera, for: AVMediaType.video, position: .front)
               else {
                   print("Unable to access back camera!")
                   return
           }
           
           do {
               let input = try AVCaptureDeviceInput(device: backCamera)

               if captureSession.canAddInput(input) {
                   captureSession.addInput(input)
               }
            
            if captureSession.canAddOutput(stillImageOutput) {
                captureSession.addOutput(stillImageOutput)
                setupLivePreview()
            }
           }
           catch let error  {
               print("Error Unable to initialize back camera:  \(error.localizedDescription)")
           }
        
        self.getCameraFrames()
           
           DispatchQueue.global(qos: .userInitiated).async {
               self.captureSession.startRunning()
           }
           
           DispatchQueue.main.async {
               self.videoPreviewLayer.frame = self.viewbaackground.bounds
           }
       }
           
    func setupLivePreview() {
        videoPreviewLayer = AVCaptureVideoPreviewLayer(session: captureSession)
        
        videoPreviewLayer.videoGravity = .resizeAspectFill
        videoPreviewLayer.connection?.videoOrientation = .portrait
        viewbaackground.layer.addSublayer(videoPreviewLayer)
        
        let overlayOval = createOverlayOval(frame: viewbaackground.frame,
        xOffset: viewbaackground.frame.midX,
        yOffset: viewbaackground.frame.midY,
        radius: 50.0)
        self.viewbaackground.addSubview(overlayOval)
        
        self.createOverlay(frame: viewOval.frame,
                                    xOffset: viewOval.frame.midX,
                                    yOffset: viewOval.frame.midY,
                                    radius: 50.0)
    }
    
    func createOverlayOval(frame: CGRect, xOffset: CGFloat, yOffset: CGFloat, radius: CGFloat) -> UIView {
        let overlayView = UIView(frame: frame)
        overlayView.backgroundColor = UIColor.black.withAlphaComponent(0.6)
        let path = CGMutablePath()
        path.addRoundedRect(in: CGRect(x: self.viewOval.frame.origin.x, y: self.viewOval.frame.origin.y,
                                           width: self.viewOval.frame.width, height: self.viewOval.frame.height ),
            cornerWidth: 0, cornerHeight: 0)

        path.addRect(CGRect(origin: .zero, size: overlayView.frame.size))
        let maskLayer = CAShapeLayer()
        maskLayer.backgroundColor = UIColor.black.cgColor
        maskLayer.path = path
        maskLayer.fillRule = .evenOdd
        overlayView.layer.mask = maskLayer
        overlayView.clipsToBounds = true

        return overlayView
    }
    
    func createOverlay(frame: CGRect, xOffset: CGFloat, yOffset: CGFloat, radius: CGFloat) {
        viewOval.backgroundColor = UIColor.black.withAlphaComponent(0.6)
        
        let pathss: CGMutablePath = {
            let path = CGMutablePath()
            path.move(to: CGPoint(x: 0, y: viewOval.frame.height/2))
            if viewOval.frame.width < viewOval.frame.height {
                path.addLine(to: CGPoint(x: 0, y: 0 + viewOval.frame.width / 2))
            }
            path.addQuadCurve(to: CGPoint(x: viewOval.frame.width / 2, y: 0), control: CGPoint(x: 0, y: 0))
            path.addQuadCurve(to: CGPoint(x: viewOval.frame.width, y: 0 + viewOval.frame.width / 2), control: CGPoint(x: viewOval.frame.width, y: 0))
            if viewOval.frame.width < viewOval.frame.height {
                path.addLine(to: CGPoint(x: viewOval.frame.width, y: viewOval.frame.height / 2))
            }
            path.addLine(to: CGPoint(x: viewOval.frame.width, y: viewOval.frame.height - viewOval.frame.width / 2))
            path.addQuadCurve(to: CGPoint(x: viewOval.frame.width / 2, y: viewOval.frame.height), control: CGPoint(x: viewOval.frame.width, y: viewOval.frame.height))
            path.addQuadCurve(to: CGPoint(x: 0, y: viewOval.frame.height -  viewOval.frame.width / 2), control: CGPoint(x: 0, y: viewOval.frame.height))
             path.addLine(to: CGPoint(x: 0, y: viewOval.frame.height / 2))
            return path
        }()

        pathss.addRect(CGRect(origin: .zero, size: viewOval.frame.size))
        let maskLayer = CAShapeLayer(layer: viewOval.layer)
        maskLayer.backgroundColor = UIColor.black.cgColor
        maskLayer.path = pathss
        maskLayer.fillRule = .evenOdd
        viewOval.layer.mask = maskLayer
        viewOval.clipsToBounds = true
    }

    //Xử lý ảnh trả về
    @available(iOS 11.0, *)
    func photoOutput(_ output: AVCapturePhotoOutput, didFinishProcessingPhoto photo: AVCapturePhoto, error: Error?) {
        guard let imageData = photo.fileDataRepresentation()
            else { return }
           
        self.captureSession.stopRunning()
        DispatchQueue.main.async {
            let viewController = SBKResultFaceVC()
            viewController.imageData = imageData
            viewController.idFront = self.idFront
            viewController.idBack = self.idBack
            viewController.URLToken = self.URLToken
            
            viewController.completionSuccessFaceData = { [weak self] data, cardImage, faceImage in
                guard let `self` = self else { return }
                self.completionSuccessFace(data, cardImage, faceImage)
            }
            
            self.navigationController?.pushViewController(viewController, animated: true)
        }
    }
    
    func resizeImage(pixelBuffer: CVPixelBuffer) -> CVPixelBuffer {
        let validateView = SBKValidateInput.shared
        let imageInput = validateView.convertCVPixelToUIImage(pixelBuffer: pixelBuffer)
        let imageOutput = validateView.cropImageFace(image: imageInput, rect: CGRect(x: imageInput.size.width / 10, y: imageInput.size.height / 10, width: imageInput.size.width * 8 / 10, height: imageInput.size.height * 8 / 10 ), scale: 1.0)
        
      return SBKValidateInput.shared.comvertUIImageToCVPixel(imageInput: imageOutput!)
    }
}

extension SBKCaptureFaceVC: AVCaptureVideoDataOutputSampleBufferDelegate {
    private func getCameraFrames() {
        self.videoDataOutput.videoSettings = [(kCVPixelBufferPixelFormatTypeKey as NSString) : NSNumber(value: kCVPixelFormatType_32BGRA)] as [String : Any]
        self.videoDataOutput.alwaysDiscardsLateVideoFrames = true
        self.videoDataOutput.setSampleBufferDelegate(self, queue: DispatchQueue(label: "camera_frame_processing_queue"))
        self.captureSession.addOutput(self.videoDataOutput)
        guard let connection = self.videoDataOutput.connection(with: AVMediaType.video), connection.isVideoOrientationSupported else { return }
        connection.videoOrientation = .portrait
    }
    
    func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        
        guard let imageFrameInput = CMSampleBufferGetImageBuffer(sampleBuffer) else {
            debugPrint("unable to get image from sample buffer")
            return
        }
        
        SBKValidateInput.shared.validateFace(imageFace: self.resizeImage(pixelBuffer: imageFrameInput)) {[weak self] data in
            guard let `self` = self else { return }
            if data {
                self.statusTake = true
                self.lbDescription.textColor = UIColor.colorFromHexa(Global.colorConfig)
                self.lbDescription.text = "Are you ready. Let's start!".localized()
            } else {
                self.statusTake = false
                self.lbDescription.textColor = UIColor.red
                self.lbDescription.text = "Position your face in the oval".localized()
            }
        }
    }
}
