//
//  RecordFace.swift
//  OCR-SDK
//
//  Created by itsol on 6/5/20.
//  Copyright © 2020 itsol. All rights reserved.
//

import UIKit
import AVFoundation
import Vision
import CoreML

class SBKRecordFace: UIViewController, AVCaptureVideoDataOutputSampleBufferDelegate {
    @IBOutlet weak var viewBackground: UIView!
    @IBOutlet weak var lbDescription: UILabel!
    @IBOutlet weak var viewOval: UIView!
    @IBOutlet weak var viewCheckStep1: UIView!
    @IBOutlet weak var viewCheckStep2: UIView!
    @IBOutlet weak var viewCheckStep3: UIView!
    @IBOutlet weak var imageDescription: UIImageView!
    
    private let captureSession = AVCaptureSession()
    private lazy var previewLayer = AVCaptureVideoPreviewLayer(session: self.captureSession)
    private let videoDataOutput = AVCaptureVideoDataOutput()
    
    private var modelDataHandler: SBKModelDataHandler? =
       SBKModelDataHandler(modelFileInfo: MobileNet.modelInfo)
    private var result: [Float]?
    private var previousInferenceTimeMs: TimeInterval = Date.distantPast.timeIntervalSince1970 * 1000
    private let delayBetweenInferencesMs: Double = 1000
    var numberPass = 0
    var numberTrue = 0
    var numberFalse = 0
    var space: Float = 0.0
    var dataImageSuccess: [Data] = []
    var checkStartRecord: Bool = false
    var timer = Timer()
    var timeRecord: Int = 0
    var checkStatusRecord: Bool = false
    var checkStep: Int = 0
    
    var idFront: String = ""
    var idBack: String = ""
    var URLToken: String = ""
    public var completionSuccessFaceRecord: (_ data: [String: Any]?, _ cardImage: Data?, _ faceImage: Data?) -> Void = {_,_,_ in}
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        self.addCameraInput()
        self.showCameraFeed()
        self.getCameraFrames()
        DispatchQueue.global(qos: .userInitiated).async {
            self.captureSession.startRunning()
        }
    }
    
    func createOverlay(frame: CGRect, xOffset: CGFloat, yOffset: CGFloat, radius: CGFloat) {
        viewOval.backgroundColor = UIColor.black.withAlphaComponent(0.6)
        
        let pathss: CGMutablePath = {
            let path = CGMutablePath()
            path.move(to: CGPoint(x: 0, y: viewOval.frame.height/2))
            if viewOval.frame.width < viewOval.frame.height {
                path.addLine(to: CGPoint(x: 0, y: 0 + viewOval.frame.width / 2))
            }
            path.addQuadCurve(to: CGPoint(x: viewOval.frame.width / 2, y: 0), control: CGPoint(x: 0, y: 0))
            path.addQuadCurve(to: CGPoint(x: viewOval.frame.width, y: 0 + viewOval.frame.width / 2), control: CGPoint(x: viewOval.frame.width, y: 0))
            if viewOval.frame.width < viewOval.frame.height {
                path.addLine(to: CGPoint(x: viewOval.frame.width, y: viewOval.frame.height / 2))
            }
            path.addLine(to: CGPoint(x: viewOval.frame.width, y: viewOval.frame.height - viewOval.frame.width / 2))
            path.addQuadCurve(to: CGPoint(x: viewOval.frame.width / 2, y: viewOval.frame.height), control: CGPoint(x: viewOval.frame.width, y: viewOval.frame.height))
            path.addQuadCurve(to: CGPoint(x: 0, y: viewOval.frame.height -  viewOval.frame.width / 2), control: CGPoint(x: 0, y: viewOval.frame.height))
            path.addLine(to: CGPoint(x: 0, y: viewOval.frame.height / 2))
            return path
        }()

        pathss.addRect(CGRect(origin: .zero, size: viewOval.frame.size))
        let maskLayer = CAShapeLayer(layer: viewOval.layer)
        maskLayer.backgroundColor = UIColor.black.cgColor
        maskLayer.path = pathss
        maskLayer.fillRule = .evenOdd
        viewOval.layer.mask = maskLayer
        viewOval.clipsToBounds = true
    }
    
    func didOutput(pixelBuffer: CVPixelBuffer, statusFace: StatusFace) {
        let currentTimeMs = Date().timeIntervalSince1970 * 1000
        guard (currentTimeMs - previousInferenceTimeMs) >= delayBetweenInferencesMs else { return }
        previousInferenceTimeMs = currentTimeMs

        // Pass the pixel buffer to TensorFlow Lite to perform inference.
        result = modelDataHandler?.runModel(onFrame: pixelBuffer)

        if self.checkStatusRecord {
            (result![0] < result![1]) ? (self.numberTrue += 1) : (self.numberFalse += 1)
            self.numberPass += 1
            DispatchQueue.main.async {
                let ciimage : CIImage = CIImage(cvPixelBuffer: pixelBuffer)
                let imageView : UIImage = SBKValidateInput.shared.convertCIToUIImage(cmage: ciimage)
                if self.result![0] < self.result![1] {
                    if statusFace == .STRAIGHTFACE && self.checkStep == 0 {
                        if self.dataImageSuccess.count == 0 {
                            self.dataImageSuccess.append(imageView.pngData()!)
                        }
                        self.lbDescription.textColor = UIColor.green
                        self.lbDescription.text = "Exactly".localized()
                        self.checkStep = 1
                        self.viewCheckStep1.backgroundColor = UIColor.colorFromHexa("#FBA02E")
                    } else if statusFace != .STRAIGHTFACE && self.checkStep == 0 {
                        self.lbDescription.textColor = UIColor.red
                        self.lbDescription.text = "Please look straight".localized()
                        self.viewCheckStep1.backgroundColor = UIColor.colorFromHexa("#FE3500")
                    } else if statusFace == .TORIGHT && self.checkStep == 1 {
                        self.lbDescription.textColor = UIColor.green
                        self.lbDescription.text = "Exactly".localized()
                        self.checkStep = 2
                        self.viewCheckStep2.backgroundColor = UIColor.colorFromHexa("#FBA02E")
                        if self.dataImageSuccess.count == 1 {
                            self.dataImageSuccess.append(imageView.pngData()!)
                        }
                    } else if statusFace != .TORIGHT && self.checkStep == 1 {
                        self.lbDescription.textColor = UIColor.red
                        self.lbDescription.text = "Please turn to the right".localized()
                        self.viewCheckStep2.backgroundColor = UIColor.colorFromHexa("#FE3500")
                    } else if statusFace == .TOLEFT && self.checkStep == 2 {
                        self.lbDescription.textColor = UIColor.green
                        self.lbDescription.text = "Exactly".localized()
                        self.checkStep = 3
                        self.viewCheckStep3.backgroundColor = UIColor.colorFromHexa("#FBA02E")
                        if self.dataImageSuccess.count == 2 {
                            self.dataImageSuccess.append(imageView.pngData()!)
                        }
                    } else if statusFace != .TOLEFT && self.checkStep == 2 {
                        self.lbDescription.textColor = UIColor.red
                        self.lbDescription.text = "Please turn to the left".localized()
                        self.viewCheckStep3.backgroundColor = UIColor.colorFromHexa("#FE3500")
                    }
                } else {
                    self.lbDescription.textColor = UIColor.red
                    self.lbDescription.text = "Incorrect face, please check!".localized()
                }
            }
        } else {
            if result![0] < result![1] {
                DispatchQueue.main.async {
                    self.checkStartRecord = true
                    self.lbDescription.textColor = UIColor.white
                    //self.lbDescription.text =  "Are you ready. Let's start!".localized()//"Bạn đã sẵn sàng. Hãy bắt đầu!"
                    if !self.checkStatusRecord {
                        self.checkStatusRecord = true
                        self.timeRecord = 0
                        self.checkStep = 0
                        self.viewCheckStep1.backgroundColor = UIColor.colorFromHexa("#333333")
                        self.viewCheckStep2.backgroundColor = UIColor.colorFromHexa("#333333")
                        self.viewCheckStep3.backgroundColor = UIColor.colorFromHexa("#333333")
                    }
                }
            } else {
                DispatchQueue.main.async {
                    self.checkStartRecord = false
                    self.lbDescription.textColor = UIColor.red
                    self.lbDescription.text = "Incorrect face, please check!".localized()
                }
            }
        }
      }
    
    init() {
        super.init(nibName: "SBKRecordFace", bundle: Bundle(for: SBKRecordFace.self))
        guard modelDataHandler != nil else {
            fatalError("Model set up failed")
        }
    }
          
    required init?(coder aDecoder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    @IBAction func onRecord(_ sender: Any) {
        if !self.checkStatusRecord {
            self.startTimer()
            self.checkStatusRecord = true
            self.timeRecord = 0
        }
    }
    
    @IBAction func onSelectBackTop(_ sender: Any) {
        for controller in self.navigationController!.viewControllers as Array {
            if controller.isKind(of: SBKCaptureCardVC.self) {
                self.navigationController!.popToViewController(controller, animated: true)
                break
            }
        }
    }
    
    func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        
        guard let frame = CMSampleBufferGetImageBuffer(sampleBuffer) else {
            debugPrint("unable to get image from sample buffer")
            return
        }
        if #available(iOS 11.0, *) {
            self.detectFace(in: self.resizeImageFace(pixelBuffer: frame))
        } else {
            // Fallback on earlier versions
        }
        
    }
    
    private func addCameraInput() {
        self.captureSession.sessionPreset = .medium
        if #available(iOS 11.1, *) {
            guard let device = AVCaptureDevice.DiscoverySession(
                deviceTypes: [.builtInWideAngleCamera, .builtInDualCamera, .builtInTrueDepthCamera],
                mediaType: .video,
                position: .front).devices.first else {
                    fatalError("No back camera device found, please make sure to run SimpleLaneDetection in an iOS device and not a simulator")
            }
            let cameraInput = try! AVCaptureDeviceInput(device: device)
            if captureSession.canAddInput(cameraInput) {
                self.captureSession.addInput(cameraInput)
            }
        }
    }
    
    private func showCameraFeed() {
        self.previewLayer.videoGravity = .resize
        self.previewLayer.frame = self.viewBackground.frame
        self.viewBackground.layer.addSublayer(self.previewLayer)
        self.lbDescription.text = "Position your face in the oval".localized()
        
        let overlayOval = createOverlayOval(frame: viewBackground.frame, xOffset: viewBackground.frame.midX, yOffset: viewBackground.frame.midY, radius: 50.0)
        self.viewBackground.addSubview(overlayOval)
        
        self.createOverlay(frame: viewOval.frame, xOffset: viewOval.frame.midX, yOffset: viewOval.frame.midY, radius: 50.0)
    }
    
    func createOverlayOval(frame: CGRect, xOffset: CGFloat, yOffset: CGFloat, radius: CGFloat) -> UIView {
        let overlayView = UIView(frame: frame)
        overlayView.backgroundColor = UIColor.black.withAlphaComponent(0.6)
        let path = CGMutablePath()
        path.addRoundedRect(in: CGRect(x: self.viewOval.frame.origin.x, y: self.viewOval.frame.origin.y, width: self.viewOval.frame.width, height: self.viewOval.frame.height ), cornerWidth: 0, cornerHeight: 0)

        path.addRect(CGRect(origin: .zero, size: overlayView.frame.size))
        let maskLayer = CAShapeLayer()
        maskLayer.backgroundColor = UIColor.black.cgColor
        maskLayer.path = path
        maskLayer.fillRule = .evenOdd
        overlayView.layer.mask = maskLayer
        overlayView.clipsToBounds = true

        return overlayView
    }
    
    private func getCameraFrames() {
        self.videoDataOutput.videoSettings = [(kCVPixelBufferPixelFormatTypeKey as NSString) : NSNumber(value: kCVPixelFormatType_32BGRA)] as [String : Any]
        self.videoDataOutput.alwaysDiscardsLateVideoFrames = true
        self.videoDataOutput.setSampleBufferDelegate(self, queue: DispatchQueue(label: "camera_frame_processing_queue"))
        self.captureSession.addOutput(self.videoDataOutput)
        guard let connection = self.videoDataOutput.connection(with: AVMediaType.video),
            connection.isVideoOrientationSupported else { return }
        connection.videoOrientation = .portrait
    }
    
    @available(iOS 11.0, *)
    private func detectFace(in image: CVPixelBuffer) {
            let faceDetectionRequest = VNDetectFaceLandmarksRequest(completionHandler: { (request: VNRequest, error: Error?) in
                DispatchQueue.main.async {
                    DispatchQueue.main.asyncAfter(deadline: .now() + 1.0) {
                    if (request.results as? [VNFaceObservation]) != nil {
                        if let results = request.results as? [VNFaceObservation], results.count > 0 {
                            
                            if results.count > 1 {
                                self.lbDescription.textColor = UIColor.red
                                self.lbDescription.text = "There are many faces in the frame".localized()
                            } else {
                                let statusString = self.checkFaceRightLeft(landmarks: results[0].landmarks!)
                                
                                self.startTimer()
                                
//                                if !self.checkStartRecord && !self.checkStatusRecord {
//                                    self.checkStartRecord = true
//                                    self.lbDescription.textColor = UIColor.white
//                                    //self.lbDescription.text =  "Are you ready. Let's start!".localized()//"Bạn đã sẵn sàng. Hãy bắt đầu!"
//                                    if !self.checkStatusRecord {
//                                        self.checkStatusRecord = true
//                                        self.timeRecord = 0
//                                        self.checkStep = 0
//                                        self.viewCheckStep1.backgroundColor = UIColor.colorFromHexa("#333333")
//                                        self.viewCheckStep2.backgroundColor = UIColor.colorFromHexa("#333333")
//                                        self.viewCheckStep3.backgroundColor = UIColor.colorFromHexa("#333333")
//                                    }
//                                }
                                
                                DispatchQueue.global().async {
                                    self.didOutput(pixelBuffer: image, statusFace: statusString)
                                }
                            }
                        } else {
                            if !self.checkStatusRecord {
                                self.checkStartRecord = false
                                self.lbDescription.textColor = UIColor.white
                                self.imageDescription.image = UIImage(named: "Scan-1@2x", in: Bundle(for: SBKTutorialVC.self), compatibleWith: nil)
                                self.lbDescription.text = "Position your face in the oval".localized()//"Đặt vị trí mặt bạn vào hình"
                            } else {
                                self.lbDescription.textColor = UIColor.red
                                self.lbDescription.text = "Incorrect face, please check!".localized()
                                
                                DispatchQueue.global().async {
                                    self.didOutput(pixelBuffer: image, statusFace: .ERROR)
                                }
                            }
                        }
                    }
                    }}
            })
            
            if #available(iOS 12.0, *) {
              // Force the revision to 2 (68-points) even on iOS 13 or greater
              // when VNDetectFaceLandmarksRequestRevision3 is available.
              faceDetectionRequest.revision = 2
            }
            
            let imageRequestHandler = VNImageRequestHandler(cvPixelBuffer: image, orientation: .leftMirrored, options: [:])
            try? imageRequestHandler.perform([faceDetectionRequest])
    }
    
    func startTimer() {
        
        switch self.checkStep {
        case 0:
            self.imageDescription.image = UIImage(named: "Scan-3@2x", in: Bundle(for: SBKTutorialVC.self), compatibleWith: nil)
            self.lbDescription.textColor = UIColor.white
            self.lbDescription.text = "Please look straight".localized()//"vui lòng nhìn thẳng"
        case 1:
            self.imageDescription.image = UIImage(named: "Scan-5@2x", in: Bundle(for: SBKTutorialVC.self), compatibleWith: nil)
            self.lbDescription.textColor = UIColor.white
            self.lbDescription.text = "Please turn to the right".localized()//"Quay sang phải"
        case 2:
            self.imageDescription.image = UIImage(named: "Scan-4@2x", in: Bundle(for: SBKTutorialVC.self), compatibleWith: nil)
            self.lbDescription.textColor = UIColor.white
            self.lbDescription.text = "Please turn to the left".localized()//"Quay sang trái"
        case 3:
            if self.dataImageSuccess.count > 2 {
                DispatchQueue.main.async {
                   let viewController = SBKResultFaceVC()
                   viewController.imageData = self.dataImageSuccess[0]
                   viewController.idFront = self.idFront
                   viewController.idBack = self.idBack
                   viewController.URLToken = self.URLToken
                   viewController.dataImageSuccess = self.dataImageSuccess
                   viewController.completionSuccessFaceData = { [weak self] data, cardImage, faceImage in
                       guard let `self` = self else { return }
                       self.completionSuccessFaceRecord(data, cardImage, faceImage)
                   }
                   self.timeRecord = 0
                    self.checkStep = 0
                   self.resetRecord()
                   self.timer.invalidate()
                   self.captureSession.stopRunning()
                   self.captureSession.removeOutput(self.videoDataOutput)
                   self.navigationController?.pushViewController(viewController, animated: true)
                }
            } else {
               self.lbDescription.textColor = UIColor.red
               self.lbDescription.text = "Unsatisfactory, please try again!".localized()//"Chưa đạt yêu cầu, hãy thử lại!"
            }
        default:
            break
        }
    }
    
    func resetRecord() {
        self.numberPass = 0
        self.numberTrue = 0
        self.numberFalse = 0
        self.space = 0.0
        self.dataImageSuccess.removeAll()
        self.checkStartRecord = false
        self.checkStatusRecord = false
    }
    
    func resizeImageFace(pixelBuffer: CVPixelBuffer) -> CVPixelBuffer {
        let validateView = SBKValidateInput.shared
        let imageInput = validateView.convertCVPixelToUIImage(pixelBuffer: pixelBuffer)
        let imageOutput = validateView.cropImageFace(image: imageInput, rect: CGRect(x: imageInput.size.width / 10, y: imageInput.size.height / 10, width: imageInput.size.width * 8 / 10, height: imageInput.size.height * 8 / 10 ), scale: 1.0)
        
        let ciimage = CIImage(image: imageOutput!)
        let tmpcontext = CIContext(options: nil)
        let cgimage =  tmpcontext.createCGImage(ciimage!, from: ciimage!.extent)
               
        return validateView.convertCGImgeToCVPixelBuffer(forImage: cgimage!)!
    }
    
    @available(iOS 11.0, *)
    func checkFaceRightLeft(landmarks: VNFaceLandmarks2D) -> StatusFace {
        let face = landmarks.faceContour?.normalizedPoints
        let nose = landmarks.nose?.normalizedPoints
        
        let faceLeftPoint = face![0]
        let faceRightPoint = face![face!.count - 1]
        let faceBottomPoint = face![5]
        let nosePoint = nose![4]
        
        let sRight = (faceBottomPoint.x - faceLeftPoint.x) * (nosePoint.y - faceLeftPoint.y) - (nosePoint.x - faceLeftPoint.x) * (faceBottomPoint.y - faceLeftPoint.y)
        let sLeft = (faceBottomPoint.x - faceRightPoint.x) * (nosePoint.y - faceRightPoint.y) - (nosePoint.x - faceRightPoint.x) * (faceBottomPoint.y - faceRightPoint.y)
        
        let sFaceRight = sqrt(sRight * sRight) / 2
        let sFaceLeft = sqrt(sLeft * sLeft) / 2
        let totalS = sFaceLeft + sFaceRight
        
        if sFaceLeft / totalS > 0.6 {
            return .TOLEFT
        } else if sFaceRight / totalS > 0.6 {
            return .TORIGHT
        } else {
            return .STRAIGHTFACE
        }
    }
}

