//
//  OCRRequest.swift
//  SSSSS
//
//  Created by itsol on 5/12/20.
//  Copyright © 2020 itsol. All rights reserved.
//

import Foundation

class SBOCRRequest: NSObject{
    
    public typealias CompletionHandle = (_ errorMess : String?, _ data: [String: Any]?) -> Void
    
    static let shared = SBOCRRequest()
    
    let certificates: [Data] = {
        let bundle = Bundle(identifier: "itsol.OCR-SDK")
        let url = bundle!.url(forResource: "objcio", withExtension: "cer")!
        let data = try! Data(contentsOf: url)
        return [data]
    }()
    
    //Call Api card front và card back
    //input: image
    //output: data[String: Any]
    func processCardFont(image: Data, pathURL: String, completion: @escaping CompletionHandle){
        
        var request = URLRequest(url: URL(string: pathURL)!)
        request.httpMethod = "POST"
        
        if !Global.header.isEmpty {
            for (key, value) in Global.header {
                request.setValue(value as? String, forHTTPHeaderField: key)
            }
        }
        
        let boundary = generateBoundaryString()

        request.setValue("multipart/form-data; boundary=\(boundary)", forHTTPHeaderField: "Content-Type")

        let imageData = image

        request.httpBody = createBodyWithParameters(filePathKey: "image", imageDataKey: imageData as NSData, boundary: boundary) as Data

        let urlSession = URLSession(configuration: URLSessionConfiguration.default, delegate: self, delegateQueue: nil)
        
        let task = urlSession.dataTask(with: request, completionHandler: { (data, _, error) -> Void in

            var errorMessage = ""

            var jsonBody = [String: Any]()

            self.processJsonBody(errorMessage: &errorMessage, withJsonBody: &jsonBody, dataResponse: data, errorResponse: error)

            guard errorMessage.count == 0 else {
                completion(errorMessage,nil)
                return
            }

            completion(nil,jsonBody)

            })
        task.resume()
    }
    
    //Call Api ProcessFace
    //input: image, idBack, idFront
    //output: data[String: Any]
    func processFace(image: Data, pathURL: String, idBack: String, idFront: String, completion: @escaping CompletionHandle){
        var params: [String: String] = [
            "id_back": idBack,
            "id_front": idFront,
        ]
        if idBack == "" {
            params.removeValue(forKey: "id_back")
        }
      
        let urlPath = URL(string: pathURL)!
        var request = URLRequest(url: urlPath)
      
      print("request====", urlPath, "==", request)
        request.httpMethod = "POST"
        
        if !Global.header.isEmpty {
            for (key, value) in Global.header {
                request.setValue(value as? String, forHTTPHeaderField: key)
            }
        }

        let boundary = generateBoundaryString()

        request.setValue("multipart/form-data; boundary=\(boundary)", forHTTPHeaderField: "Content-Type")

        let imageData = image

        request.httpBody = createBodyWithParameters(parameters: params, filePathKey: "image_general", imageDataKey: imageData as NSData, boundary: boundary) as Data
        
        let urlSession = URLSession(configuration: URLSessionConfiguration.default, delegate: self, delegateQueue: nil)
        
        let task = urlSession.dataTask(with: request, completionHandler: { (data, respone, error) -> Void in

            var errorMessage = ""

            var jsonBody = [String: Any]()
            
            self.processJsonBody(errorMessage: &errorMessage, withJsonBody: &jsonBody, dataResponse: data, errorResponse: error)

            guard errorMessage.count == 0 else {
                completion(errorMessage,nil)
                return
            }

            completion(nil,jsonBody)

            })
        task.resume()
    }
    
    func createBodyWithParameters(parameters: [String: String] = [:], filePathKey: String?, imageDataKey: NSData, boundary: String) -> NSData {
        let body = NSMutableData();
        for (key, value) in parameters {
            body.appendString(string: "--\(boundary)\r\n")
            body.appendString(string: "Content-Disposition: form-data; name=\"\(key)\"\r\n\r\n")
            body.appendString(string: "\(value)\r\n")
        }
        let face = "face.jpg"
        let card1 = "cardFront.jpg"
        let card2 = "cardBack.jpg"
        let filePathKeyCard1 = "image_card1"
        let filePathKeyCard2 = "image_card2"
        let mimetype = "image/jpg"

        body.appendString(string: "--\(boundary)\r\n")
        body.appendString(string: "Content-Disposition: form-data; name=\"\(filePathKey!)\"; filename=\"\(face)\"\r\n")
        body.appendString(string: "Content-Type: \(mimetype)\r\n\r\n")
        body.append(imageDataKey as Data)
        body.appendString(string: "\r\n")
        
        if Global.imageCard1 != nil {
            body.appendString(string: "--\(boundary)\r\n")
            body.appendString(string: "Content-Disposition: form-data; name=\"\(filePathKeyCard1)\"; filename=\"\(card1)\"\r\n")
            body.appendString(string: "Content-Type: \(mimetype)\r\n\r\n")
            body.append(Global.imageCard1!)
            body.appendString(string: "\r\n")
        }
        
        if Global.imageCard2 != nil {
            body.appendString(string: "--\(boundary)\r\n")
            body.appendString(string: "Content-Disposition: form-data; name=\"\(filePathKeyCard2)\"; filename=\"\(card2)\"\r\n")
            body.appendString(string: "Content-Type: \(mimetype)\r\n\r\n")
            body.append(Global.imageCard2!)
            body.appendString(string: "\r\n")
        }

        body.appendString(string: "--\(boundary)--\r\n")

        return body
    }
    
    func generateBoundaryString() -> String {
        return "Boundary-\(NSUUID().uuidString)"
    }
    
    private func processJsonBody (errorMessage errorMess: inout String,
                                  withJsonBody jsonResult: inout [String: Any],
                                  dataResponse data: Data?,
                                  errorResponse error: Error?  ) {
        
        guard error == nil else {
            errorMess = "\((error! as NSError).code)"
            return
        }
        
        guard data != nil else {
            errorMess = "Error: Did not receive data"
            return
        }
        
        //let str = String(decoding: data!, as: UTF8.self)
        
        guard let _jsonResult = try? JSONSerialization.jsonObject(with: data!, options: JSONSerialization.ReadingOptions.mutableContainers) as? [String: Any] else {
            errorMess = "Error: data can not convert to jsonObject"
            return
        }
        
        jsonResult = _jsonResult
    }
}

extension SBOCRRequest: URLSessionDelegate {
    
    func urlSession(_ session: URLSession, didReceive challenge: URLAuthenticationChallenge, completionHandler: @escaping (URLSession.AuthChallengeDisposition, URLCredential?) -> Void) {
        print("urlsession====", Global.useFileCer)
        if let trust = challenge.protectionSpace.serverTrust, SecTrustGetCertificateCount(trust) > 0, Global.useFileCer {
            if let certificate = SecTrustGetCertificateAtIndex(trust, 0) {
                let data = SecCertificateCopyData(certificate) as Data
                if certificates.contains(data) {
                    completionHandler(.useCredential, URLCredential(trust: trust))
                    return
                }
            }
        }
        completionHandler(.cancelAuthenticationChallenge, nil)
    }
}

extension NSMutableData {

    func appendString(string: String) {
        let data = string.data(using: String.Encoding.utf8, allowLossyConversion: true)
        append(data!)
    }
}
